#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "common.h"
#include "crc.h"

#define GOLD_ACTION_SAVE  1
#define GOLD_ACTION_CHECK 2

#define GOLD_TYPE_CRC     1
#define GOLD_TYPE_CONTENT 2

char * app_name = 0;

int      support_opt_scale_size = 1;
uint32_t support_opt_gold_surface_type_mask = 0;

#if !defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
char const * opt_stream_file = 0;
#endif
int          opt_repeat_count = 1;
float        opt_scale_size = 1.0;
int          opt_gold_action = 0;
uint32_t     opt_gold_every = 0;
int          opt_gold_type = 0;
uint32_t     opt_gold_surface_type = 0;
int          opt_gold_file_per_frame = 0;
char const * opt_gold_path = 0;
int          opt_gold_bmp = 0;
int          opt_gold_raw = 0;

void *                  trace_stream_handle;
Traceplay_Stream_Conf * trace_stream_conf_ptr;
Traceplay_Cmds        * trace_cmds_ptr;

void common_init()
{
    if (!app_name) {
        printf("ERROR: App didn't set app_name\n");
        common_exit_error();
    }
}

void common_print_help()
{
    printf("    -h/-?/-help/--help: Print help.\n");
#if !defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
    printf("    --stream <filename>: Load stream to play from <filename>\n");
#endif
    printf("    --repeat <count>: Repeats stream <count> times (0==infinite)\n");
    if (support_opt_scale_size) {
        printf("    --scale-size <scale>: Scale output surface size by <scale> (a float).\n");
    }
    if (support_opt_gold_surface_type_mask) {
        printf("    --gold-action <action>: Enable gold generation/checking, where <action> is:\n");
        printf("        save: Save golds\n");
        printf("        check: Check against pre-existing golds\n");

        printf("    --gold-every <n>: Save/check golds only every <n> frames\n");

        printf("    --gold-type <type>: Specify the type of gold to save/check:\n");
        printf("        crc: CRC the surface, and use that as the data\n");
        printf("        content: Save/check the entire surface image\n");

        printf("    --gold-surface-type <type>: Specify the type of surface to save/check:\n");
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER)) {
            printf("        vid-dec: VdpVideoSurface using VdpVideoSurfaceGetBitsYCbCr (decode order)\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER)) {
            printf("        vid-disp: VdpVideoSurface using VdpVideoSurfaceGetBitsYCbCr (display order)\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_OUT)) {
            printf("        out: VdpOutputSurface after post-processing using VdpOutputSurfaceGetBitsNative\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_TFP_OUT)) {
            printf("        gl-tfp-out: GL texture from GL_EXT_texture_from_pixmap\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_INTEROP_VID)) {
            printf("        gl-interop-vid: GL texture from glVDPAUMapSurfacesNV for a VdpVideoSurface (display order)\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_INTEROP_OUT)) {
            printf("        gl-interop-out: GL texture from glVDPAUMapSurfacesNV for a VdpOutputSurface\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_WIN)) {
            printf("        gl-win: Screengrab of the entire GL window using glReadPixels\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID)) {
            printf("        cuda-driver-interop-vid: CUDA array from cuGraphicsSubResourceGetMappedArray for a VdpVideoSurface (display order)\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT)) {
            printf("        cuda-driver-interop-out: CUDA array from cuGraphicsSubResourceGetMappedArray for a VdpOutputSurface\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID)) {
            printf("        cuda-toolkit-interop-vid: CUDA array from cudaGraphicsSubResourceGetMappedArray for a VdpVideoSurface (display order)\n");
        }
        if (support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT)) {
            printf("        cuda-toolkit-interop-out: CUDA array from cudaGraphicsSubResourceGetMappedArray for a VdpOutputSurface\n");
        }
        printf("    --gold-file-per-frame: Store each frame's gold in a separate frame,\n");
        printf("        otherwise, each frame's gold will be concatenated into a single file.\n");
        printf("    --gold-path <path>: The gold directory or filename\n");
        printf("        Without --gold-file-per-frame, this is the name of the single gold file.\n");
        printf("        With --gold-file-per-frame, gold files will be named internally and written to this directory.\n");
        printf("    --gold-bmp: Emit a BMP header on golds; legal only for RGB golds\n");
        printf("    --gold-raw: Don't prepend a 'tagline' to golds; just emit raw RGB/BMP/NV12 data. --gold-bmp is legal with this.\n");
    }
}

int common_parse_arg(int * pargc, char * * * pargv)
{
#if !defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
    if (!strcmp((*pargv)[0], "--stream")) {
        if (*pargc < 2) {
            printf("ERROR: --stream requires an argument <filename>\n");
            return COMMON_PARSE_ARG_ERROR;
        }
        (*pargc)--;
        (*pargv)++; 
        opt_stream_file = (*pargv)[0];
        return COMMON_PARSE_ARG_OK;
    }
#endif
    if (!strcmp((*pargv)[0], "--repeat")) {
        if (*pargc < 2) {
            printf("ERROR: --repeat requires an argument <count>\n");
            return COMMON_PARSE_ARG_ERROR;
        }
        (*pargc)--;
        (*pargv)++; 
        opt_repeat_count = atoi((*pargv)[0]);
        return COMMON_PARSE_ARG_OK;
    }
    if (support_opt_scale_size && !strcmp((*pargv)[0], "--scale-size")) {
        if ((*pargc) < 2) {
            printf("ERROR: --scale-size requires an argument <scale>\n");
            return COMMON_PARSE_ARG_ERROR;
        }
        (*pargc)--;
        (*pargv)++; 
        opt_scale_size = atof((*pargv)[0]);
        return COMMON_PARSE_ARG_OK;
    }
    if (
        !strcmp((*pargv)[0], "-h") ||
        !strcmp((*pargv)[0], "-help") ||
        !strcmp((*pargv)[0], "--help") ||
        !strcmp((*pargv)[0], "-?")
    ) {
        // This should trigger help printing
        return COMMON_PARSE_ARG_ERROR;
    }
    if (support_opt_gold_surface_type_mask) {
        if (!strcmp((*pargv)[0], "--gold-action")) {
            if (*pargc < 2) {
                printf("ERROR: --gold-action requires an argument <action>\n");
                return COMMON_PARSE_ARG_ERROR;
            }
            (*pargc)--;
            (*pargv)++; 
            if (!strcmp((*pargv)[0], "save")) {
                opt_gold_action = GOLD_ACTION_SAVE;
            }
            else if (!strcmp((*pargv)[0], "check")) {
                opt_gold_action = GOLD_ACTION_CHECK;
            }
            else {
                printf("ERROR: Unknown --gold-action '%s'\n", (*pargv)[0]);
                return COMMON_PARSE_ARG_ERROR;
            }
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-every")) {
            if (*pargc < 2) {
                printf("ERROR: --gold-every requires an argument <count>\n");
                return COMMON_PARSE_ARG_ERROR;
            }
            (*pargc)--;
            (*pargv)++; 
            opt_gold_every = atoi((*pargv)[0]);
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-type")) {
            if (*pargc < 2) {
                printf("ERROR: --gold-type requires an argument <type>\n");
                return COMMON_PARSE_ARG_ERROR;
            }
            (*pargc)--;
            (*pargv)++; 
            if (!strcmp((*pargv)[0], "crc")) {
                opt_gold_type = GOLD_TYPE_CRC;
            }
            else if (!strcmp((*pargv)[0], "content")) {
                opt_gold_type = GOLD_TYPE_CONTENT;
            }
            else {
                printf("ERROR: Unknown --gold-type '%s'\n", (*pargv)[0]);
                return COMMON_PARSE_ARG_ERROR;
            }
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-surface-type")) {
            if (*pargc < 2) {
                printf("ERROR: --gold-surface_type requires an argument <type>\n");
                return COMMON_PARSE_ARG_ERROR;
            }
            (*pargc)--;
            (*pargv)++; 
            if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER)) && !strcmp((*pargv)[0], "vid-dec")) {
                opt_gold_surface_type = SURFACE_TYPE_VDP_VID_DEC_ORDER;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER)) && !strcmp((*pargv)[0], "vid-disp")) {
                opt_gold_surface_type = SURFACE_TYPE_VDP_VID_DISP_ORDER;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_VDP_OUT)) && !strcmp((*pargv)[0], "out")) {
                opt_gold_surface_type = SURFACE_TYPE_VDP_OUT;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_TFP_OUT)) && !strcmp((*pargv)[0], "gl-tfp-out")) {
                opt_gold_surface_type = SURFACE_TYPE_GL_TFP_OUT;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_INTEROP_VID)) && !strcmp((*pargv)[0], "gl-interop-vid")) {
                opt_gold_surface_type = SURFACE_TYPE_GL_INTEROP_VID;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_INTEROP_OUT)) && !strcmp((*pargv)[0], "gl-interop-out")) {
                opt_gold_surface_type = SURFACE_TYPE_GL_INTEROP_OUT;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_GL_WIN)) && !strcmp((*pargv)[0], "gl-win")) {
                opt_gold_surface_type = SURFACE_TYPE_GL_WIN;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID)) && !strcmp((*pargv)[0], "cuda-driver-interop-vid")) {
                opt_gold_surface_type = SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT)) && !strcmp((*pargv)[0], "cuda-driver-interop-out")) {
                opt_gold_surface_type = SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID)) && !strcmp((*pargv)[0], "cuda-toolkit-interop-vid")) {
                opt_gold_surface_type = SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID;
            }
            else if ((support_opt_gold_surface_type_mask & (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT)) && !strcmp((*pargv)[0], "cuda-toolkit-interop-out")) {
                opt_gold_surface_type = SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT;
            }
            else {
                printf("ERROR: Unknown --gold-surface-type '%s'\n", (*pargv)[0]);
                return COMMON_PARSE_ARG_ERROR;
            }
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-file-per-frame")) {
            opt_gold_file_per_frame = 1;
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-path")) {
            if (*pargc < 2) {
                printf("ERROR: --gold-path requires an argument <path>\n");
                return COMMON_PARSE_ARG_ERROR;
            }
            (*pargc)--;
            (*pargv)++; 
            opt_gold_path = (*pargv)[0];
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-bmp")) {
            opt_gold_bmp = 1;
            return COMMON_PARSE_ARG_OK;
        }

        if (!strcmp((*pargv)[0], "--gold-raw")) {
            opt_gold_raw = 1;
            return COMMON_PARSE_ARG_OK;
        }
    }

    return COMMON_PARSE_ARG_NOT_PARSED;
}

int common_check_args()
{
#if !defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
    if (!opt_stream_file) {
        printf("ERROR: --stream not specified; it must be\n");
        return COMMON_CHECK_ARGS_ERROR;
    }
#endif
    if (opt_gold_action) {
        if (!opt_gold_type) {
            printf("ERROR: --gold-action specified, but --gold-type was not\n");
            return COMMON_CHECK_ARGS_ERROR;
        }
        if (!opt_gold_surface_type) {
            printf("ERROR: --gold-action specified, but --gold-surface-type was not\n");
            return COMMON_CHECK_ARGS_ERROR;
        }
        if (!opt_gold_path) {
            printf("ERROR: --gold-action specified, but --gold-path was not\n");
            return COMMON_CHECK_ARGS_ERROR;
        }
        if (opt_gold_file_per_frame) {
            struct stat st;
            if (stat(opt_gold_path, &st)) {
                printf("ERROR: ----gold-file-per-frame specified, but can't stat path '%s'\n", opt_gold_path);
                return COMMON_CHECK_ARGS_ERROR;
            }
            if (!S_ISDIR(st.st_mode)) {
                printf("ERROR: ----gold-file-per-frame specified, but path '%s' is not a directory\n", opt_gold_path);
                return COMMON_CHECK_ARGS_ERROR;
            }
        }
        if (opt_gold_bmp) {
            switch (opt_gold_surface_type) {
            case SURFACE_TYPE_VDP_OUT:
            case SURFACE_TYPE_GL_TFP_OUT:
            case SURFACE_TYPE_GL_INTEROP_OUT:
            case SURFACE_TYPE_GL_WIN:
            case SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT:
            case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT:
                break;
            case SURFACE_TYPE_VDP_VID_DEC_ORDER:
            case SURFACE_TYPE_VDP_VID_DISP_ORDER:
            case SURFACE_TYPE_GL_INTEROP_VID:
            case SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID:
            case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID:
            default:
                printf("ERROR: --gold-bmp specified, but --gold-surface-type is not an RGB type\n");
                return COMMON_CHECK_ARGS_ERROR;
            }
            switch (opt_gold_type) {
            case GOLD_TYPE_CONTENT:
                break;
            case GOLD_TYPE_CRC:
            default:
                printf("ERROR: --gold-bmp specified, but --gold--type is not 'content'\n");
                return COMMON_CHECK_ARGS_ERROR;
            }
        }
    }

    return COMMON_CHECK_ARGS_OK;
}

#if defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
extern Traceplay_Stream_Conf traceplay_stream_conf;
extern Traceplay_Cmds *traceplay_cmds;
#endif

void common_load_stream()
{
#if !defined(TRACEPLAY_DIRECTLY_LINK_STREAM)
    trace_stream_handle = dlopen(opt_stream_file, RTLD_NOW | RTLD_LOCAL);
    if (!trace_stream_handle) {
        printf("ERROR at %s:%d: Could not load stream '%s': dlopen failed: %s\n", __FILE__, __LINE__, opt_stream_file, dlerror());
        common_exit_error();
    }

    trace_stream_conf_ptr = (Traceplay_Stream_Conf *)dlsym(trace_stream_handle, "traceplay_stream_conf");
    if (!trace_stream_conf_ptr) {
        printf("ERROR at %s:%d: Could not load stream '%s': dlsym(traceplay_stream_conf) failed: %s\n", __FILE__, __LINE__, opt_stream_file, dlerror());
        common_exit_error();
    }

    Traceplay_Cmds ** trace_cmds_ptr_ptr;
    trace_cmds_ptr_ptr = (Traceplay_Cmds **)dlsym(trace_stream_handle, "traceplay_cmds");
    if (!trace_cmds_ptr_ptr) {
        printf("ERROR at %s:%d: Could not load stream '%s': dlsym(trace_cmds_ptr) failed: %s\n", __FILE__, __LINE__, opt_stream_file, dlerror());
        common_exit_error();
    }
    trace_cmds_ptr = *trace_cmds_ptr_ptr;
#else
    trace_stream_conf_ptr = &traceplay_stream_conf;
    trace_cmds_ptr = traceplay_cmds;
#endif

    // Hack to force CUDA CSC kernel to do scaling too.
    trace_stream_conf_ptr->out_width *= opt_scale_size;
    trace_stream_conf_ptr->out_height *= opt_scale_size;
}

int forced_exit_failure = 0;

void common_force_failure_but_dont_exit()
{
    forced_exit_failure = 1;
}

void common_exit_ok()
{
    if (!app_name) {
        common_exit_error();
    }
    if (forced_exit_failure) {
        common_exit_error();
    }
    printf("&&&& %s test PASSED\n", app_name);
    exit(0);
}

void common_exit_error()
{
    printf("&&&& %s test FAILED\n", app_name ? app_name : "unknown");
    exit(1);
}

char const * fp_gold_mode;
FILE *fp_gold;

VdpVideoSurfaceGetBitsYCbCr *   vdp_video_surface_get_bits_y_cb_cr;
VdpOutputSurfaceGetBitsNative * vdp_output_surface_get_bits_native;

uint32_t surface_buffer_size;
uint8_t *scratch_surface_buffer;
uint8_t *surface_buffer;
uint8_t *gold_buffer;

void common_playback_init(
    VdpDevice           vdp_device,
    VdpGetProcAddress * vdp_get_proc_address,
    int                 need_scratch_surface_buffer
)
{
    if (opt_gold_action) {
        fp_gold_mode = (opt_gold_action == GOLD_ACTION_SAVE) ? "wb" : "rb";
        if (!opt_gold_file_per_frame) {
            fp_gold = fopen(opt_gold_path, fp_gold_mode);
            if (!fp_gold) {
                printf("ERROR at %s:%d: Could not open gold file '%s' with mode '%s': %s\n", __FILE__, __LINE__, opt_gold_path, fp_gold_mode, strerror(errno));
                common_exit_error();
            }
        }

        VdpStatus vdp_st;
        switch (opt_gold_surface_type) {
        case SURFACE_TYPE_VDP_VID_DEC_ORDER:
        case SURFACE_TYPE_VDP_VID_DISP_ORDER:
            vdp_st = vdp_get_proc_address(
                vdp_device,
                VDP_FUNC_ID_VIDEO_SURFACE_GET_BITS_Y_CB_CR,
                (void **)&vdp_video_surface_get_bits_y_cb_cr
            );
            CHECK_ST
            break;
        case SURFACE_TYPE_VDP_OUT:
            vdp_st = vdp_get_proc_address(
                vdp_device,
                VDP_FUNC_ID_OUTPUT_SURFACE_GET_BITS_NATIVE,
                (void **)&vdp_output_surface_get_bits_native
            );
            CHECK_ST
            break;
        default:
            break;
        }

        switch (opt_gold_surface_type) {
        case SURFACE_TYPE_VDP_VID_DEC_ORDER:
        case SURFACE_TYPE_VDP_VID_DISP_ORDER:
        case SURFACE_TYPE_GL_INTEROP_VID:
        case SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID:
        case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID:
            // FIXME: Assumes 4:2:0; 4:2:2 isn't supported here...
            surface_buffer_size = (trace_stream_conf_ptr->vid_width * trace_stream_conf_ptr->vid_height * 3) / 2;
            break;
        case SURFACE_TYPE_VDP_OUT:
        case SURFACE_TYPE_GL_TFP_OUT:
        case SURFACE_TYPE_GL_INTEROP_OUT:
        case SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT:
        case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT:
            // FIXME: Assumes 4:4:4 ARGB, although that's all that VDPAU supports
            surface_buffer_size = trace_stream_conf_ptr->out_width * trace_stream_conf_ptr->out_height * 4;
            break;
        case SURFACE_TYPE_GL_WIN:
            if (!surface_buffer_size) {
                printf("ERROR at %s:%d: surface_buffer_size not set by application\n", __FILE__, __LINE__);
                common_exit_error();
            }
            break;
        default:
            printf("ERROR at %s:%d: gold surface type %d not supported\n", __FILE__, __LINE__, opt_gold_surface_type);
            common_exit_error();
            break;
        }
        if (surface_buffer_size) {
            surface_buffer = (uint8_t *)malloc(surface_buffer_size);
            CHECK_PTR(surface_buffer)
            if (need_scratch_surface_buffer) {
                scratch_surface_buffer = (uint8_t *)malloc(surface_buffer_size);
                CHECK_PTR(scratch_surface_buffer)
            }
            gold_buffer = (uint8_t *)malloc(surface_buffer_size);
            CHECK_PTR(gold_buffer)
        }
    }
}

void common_do_gold_vdp_video_surface(uint32_t frame_number, VdpVideoSurface surface)
{
    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    uint32_t destination_pitches[] = {
        trace_stream_conf_ptr->vid_width,
        trace_stream_conf_ptr->vid_width
    };
    uint32_t destination_sizes[] = {
        destination_pitches[0] * trace_stream_conf_ptr->vid_height,
        destination_pitches[1] * trace_stream_conf_ptr->vid_height,
    };
    void * destination_data[] = {
        surface_buffer,
        surface_buffer + destination_sizes[0],
    };

    VdpStatus vdp_st = vdp_video_surface_get_bits_y_cb_cr(
        surface,
        VDP_YCBCR_FORMAT_NV12,
        destination_data,
        destination_pitches
    );
    CHECK_ST

    common_do_gold_buffer(frame_number, "nv12", trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, surface_buffer, surface_buffer_size);
}

void common_do_gold_vdp_output_surface(uint32_t frame_number, VdpOutputSurface surface)
{
    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    uint32_t destination_pitches[] = {
        trace_stream_conf_ptr->out_width * 4,
    };
    void * destination_data[] = {
        surface_buffer,
    };

    VdpStatus vdp_st = vdp_output_surface_get_bits_native(
        surface,
        0,
        destination_data,
        destination_pitches
    );
    CHECK_ST

    common_do_gold_buffer_argb_to_rgb(frame_number, trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height, surface_buffer, surface_buffer_size, 0);
}

crc_t calccrc(uint8_t const * buffer, size_t buffer_size_bytes)
{
    crc_t crc = crc_init();
    crc = crc_update(crc, buffer, buffer_size_bytes);
    crc = crc_finalize(crc);
    return crc;
}

void common_do_gold_buffer_argb_to_rgb(uint32_t frame_number, uint32_t width, uint32_t height, uint8_t * buffer, size_t buffer_size_bytes, int vinvert)
{
    int s, d;
    for (s = 0, d = 0; s < buffer_size_bytes; s += 4, d += 3) {
        buffer[d + 0] = buffer[s + 0];
        buffer[d + 1] = buffer[s + 1];
        buffer[d + 2] = buffer[s + 2];
    }

    common_do_gold_buffer(frame_number, vinvert ? "rgb-vinvert" : "rgb", width, height, buffer, buffer_size_bytes * 3 / 4);

}

void common_do_gold_buffer(uint32_t frame_number, char const * format, uint32_t width, uint32_t height, uint8_t const * buffer, size_t buffer_size_bytes)
{
    if (opt_gold_file_per_frame) {
        char const * extension = 0;
        switch (opt_gold_surface_type) {
        case SURFACE_TYPE_VDP_OUT:
        case SURFACE_TYPE_GL_TFP_OUT:
        case SURFACE_TYPE_GL_INTEROP_OUT:
        case SURFACE_TYPE_GL_WIN:
        case SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT:
        case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT:
            if (opt_gold_raw && opt_gold_bmp) {
                extension = "bmp";
            }
            break;
        case SURFACE_TYPE_VDP_VID_DEC_ORDER:
        case SURFACE_TYPE_VDP_VID_DISP_ORDER:
        case SURFACE_TYPE_GL_INTEROP_VID:
        case SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID:
        case SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID:
        default:
            break;
        }
        if (!extension && opt_gold_raw) {
            extension = format;
        }
        if (!extension) {
            extension = "tpgold";
        }

        char path[FILENAME_MAX];
        sprintf(path, "%s/frame_%06d.%s", opt_gold_path, frame_number, extension);
        fp_gold = fopen(path, fp_gold_mode);
        if (!fp_gold) {
            printf("ERROR at %s:%d: Could not open gold file '%s' with mode '%s': %s\n", __FILE__, __LINE__, path, fp_gold_mode, strerror(errno));
            common_exit_error();
        }
    }

    char const * type;
    if (opt_gold_type == GOLD_TYPE_CRC) {
        type = "crc";
    }
    else {
        type = "content";
    }

    char crcbuffer[256];
    if (opt_gold_type == GOLD_TYPE_CRC) {
        sprintf(crcbuffer, "  crc=0x%08x\n", calccrc(buffer, buffer_size_bytes));
        buffer = (uint8_t *)crcbuffer;
        buffer_size_bytes = strlen(crcbuffer);
    }

    uint8_t * content_header = 0;
    uint32_t content_header_size = 0;

    if (opt_gold_bmp) {
        static uint8_t bmp_header[54];

        uint32_t bmp_size = 54 + buffer_size_bytes;
        int32_t  bmp_height = height;
        // "rgb" means top-to-bottom. "rgb-invert" means bottom-to-top.
        //  bmp stores images bottom-to-top, except when height is negative.
        if (!strcmp(format, "rgb")) {
            bmp_height *= -1;
        }

        memset(bmp_header, 0, sizeof(bmp_header));

        bmp_header[0] = 'B';                      // Magic
        bmp_header[1] = 'M';
        bmp_header[2] = bmp_size;                 // File size
        bmp_header[3] = bmp_size >> 8;
        bmp_header[4] = bmp_size >> 16;
        bmp_header[5] = bmp_size >> 24;
        bmp_header[10] = 54;                      // Image data offset
        bmp_header[14] = 40;                      // Size of remaining header
        bmp_header[18] = width;                   // Image width
        bmp_header[19] = width >> 8;
        bmp_header[20] = width >> 16;
        bmp_header[21] = width >> 24;             // Image height
        bmp_header[22] = bmp_height;
        bmp_header[23] = bmp_height >> 8;
        bmp_header[24] = bmp_height >> 16;
        bmp_header[25] = bmp_height >> 24;
        bmp_header[26] = 1;                       // Number of planes
        bmp_header[28] = 24;                      // Bits per pixel
        bmp_header[34] = buffer_size_bytes;       // Image data size in bytes
        bmp_header[35] = buffer_size_bytes >> 8;
        bmp_header[36] = buffer_size_bytes >> 16;
        bmp_header[37] = buffer_size_bytes >> 24;
        bmp_header[38] = 96;                      // Horizontal DPI
        bmp_header[42] = 96;                      // Vertical DPI

        content_header = bmp_header;
        content_header_size = sizeof(bmp_header);

        format = "bmp";
    }

    char tagline[256];
    sprintf(tagline, "traceplay gold %s %s %u %u content_size=%u\n", type, format, width, height, (unsigned int)(buffer_size_bytes + content_header_size));

    if (opt_gold_action == GOLD_ACTION_SAVE) {
        if (!opt_gold_raw) {
            fwrite(tagline, 1, strlen(tagline), fp_gold);
        }
        if (content_header) {
            fwrite(content_header, 1, content_header_size, fp_gold);
        }
        fwrite(buffer, 1, buffer_size_bytes, fp_gold);
        fflush(fp_gold);
    }
    else {
        if (!opt_gold_raw) {
            char tagline_gold[256];
            if (fgets(tagline_gold, sizeof(tagline_gold), fp_gold) != tagline_gold) {
                printf("ERROR at %s:%d: Could not read gold file: %s\n", __FILE__, __LINE__, strerror(errno));
                common_exit_error();
            }
            if (strcmp(tagline, tagline_gold)) {
                printf("ERROR at %s:%d: Frame %d's gold tagline did not match:\n", __FILE__, __LINE__, frame_number);
                printf("    Calculated: %s", tagline);
                printf("    Gold:       %s", tagline_gold);
                common_exit_error();
            }
        }
        if (content_header) {
            if (fread(gold_buffer, 1, content_header_size, fp_gold) != content_header_size) {
                printf("ERROR at %s:%d: Could not read gold file: %s\n", __FILE__, __LINE__, strerror(errno));
                common_exit_error();
            }
            if (memcmp(content_header, gold_buffer, content_header_size)) {
                printf("ERROR at %s:%d: Frame %d's gold mismatched (in header data):\n", __FILE__, __LINE__, frame_number);
                common_force_failure_but_dont_exit();
            }
        }
        if (fread(gold_buffer, 1, buffer_size_bytes, fp_gold) != buffer_size_bytes) {
            printf("ERROR at %s:%d: Could not read gold file: %s\n", __FILE__, __LINE__, strerror(errno));
            common_exit_error();
        }
        if (memcmp(buffer, gold_buffer, buffer_size_bytes)) {
            printf("ERROR at %s:%d: Frame %d's gold mismatched:\n", __FILE__, __LINE__, frame_number);
            if (opt_gold_type == GOLD_TYPE_CRC) {
                printf("    Calculated: %s", buffer);
                printf("    Gold:       %s", gold_buffer);
            }
            else {
                printf("    Calculated CRC: 0x%08x\n", calccrc(buffer, buffer_size_bytes));
                printf("    Gold CRC:       0x%08x\n", calccrc(gold_buffer, buffer_size_bytes));
            }
            common_force_failure_but_dont_exit();
        }
    }

    if (opt_gold_file_per_frame) {
        if (fclose(fp_gold)) {
            printf("ERROR at %s:%d: Could not close gold file: %s\n", __FILE__, __LINE__, strerror(errno));
            common_exit_error();
        }
        fp_gold = 0;
    }
}

void common_playback_fini()
{
    if (fp_gold) {
        if (fclose(fp_gold)) {
            printf("ERROR at %s:%d: Could not close gold file: %s\n", __FILE__, __LINE__, strerror(errno));
            common_exit_error();
        }
        fp_gold = 0;
    }
    if (gold_buffer) {
        free(gold_buffer);
        gold_buffer = 0;
    }
    if (scratch_surface_buffer) {
        free(scratch_surface_buffer);
        scratch_surface_buffer = 0;
    }
    if (surface_buffer) {
        free(surface_buffer);
        surface_buffer = 0;
    }
}

void common_nv12x2_to_nv12(uint32_t width, uint32_t height, uint8_t const * src, uint8_t * dest)
{
    uint32_t pitch_y = width;
    uint32_t count_y = height / 2;
    uint32_t size_y  = pitch_y * count_y;
    uint32_t pitch_uv = width;
    uint32_t count_uv = height / 4;
    uint32_t size_uv  = pitch_uv * count_uv;

    uint32_t offset_in_y_top  = 0;
    uint32_t offset_in_y_bot  = offset_in_y_top  + size_y;
    uint32_t offset_in_uv_top = offset_in_y_bot  + size_y;
    uint32_t offset_in_uv_bot = offset_in_uv_top + size_uv;

    uint32_t offset_out_y  = 0;
    uint32_t offset_out_uv = offset_out_y + (2 * size_y);

    int i;

    for (i = 0; i < count_y; ++i) {
        memcpy(dest + offset_out_y  + (pitch_y  * ((2 * i) + 0)), src + offset_in_y_top  + (pitch_y  * i), pitch_y );
        memcpy(dest + offset_out_y  + (pitch_y  * ((2 * i) + 1)), src + offset_in_y_bot  + (pitch_y  * i), pitch_y );
    }

    for (i = 0; i < count_uv; ++i) {
        memcpy(dest + offset_out_uv + (pitch_uv * ((2 * i) + 0)), src + offset_in_uv_top + (pitch_uv * i), pitch_uv);
        memcpy(dest + offset_out_uv + (pitch_uv * ((2 * i) + 1)), src + offset_in_uv_bot + (pitch_uv * i), pitch_uv);
    }
}

